"use strict";
/// <reference types="node" />
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const events_1 = require("events");
const fs_1 = tslib_1.__importDefault(require("fs"));
const IORedis = require("ioredis");
const path_1 = tslib_1.__importDefault(require("path"));
const lua = process.env.DEBUG_SLIDING_WINDOW_RATELIMITER_LUA
    ? fs_1.default.readFileSync(path_1.default.join(__dirname, '../src/sliding-window-rate-limiter.lua'), 'utf8')
    : fs_1.default.readFileSync(path_1.default.join(__dirname, '../lib/sliding-window-rate-limiter.min.lua'), 'utf8');
class RedisSlidingWindowRateLimiter extends events_1.EventEmitter {
    constructor(options = {}) {
        super();
        this.options = options;
        this.interval = Number(options.interval) || 60;
        this.operationTimeout = options.operationTimeout || 0;
        if (!options.redis || typeof options.redis === 'string') {
            this.redis = new IORedis({
                host: options.redis,
                retryStrategy: (_times) => 1000,
                maxRetriesPerRequest: 1
            });
        }
        else {
            this.redis = options.redis;
        }
        this.redis.defineCommand('limiter', {
            lua,
            numberOfKeys: 1
        });
    }
    check(key, limit) {
        return this.limiter(key, 0 /* Check */, this.interval, limit, 0);
    }
    reserve(key, limit) {
        return this.limiter(key, 1 /* Reserve */, this.interval, limit, 0);
    }
    cancel(key, ts) {
        return this.limiter(key, 2 /* Cancel */, this.interval, 0, ts);
    }
    destroy() {
        if (!this.options.redis || typeof this.options.redis === 'string') {
            try {
                this.redis.quit().catch(() => {
                    this.redis.disconnect();
                });
            }
            catch (e) {
                this.redis.disconnect();
            }
        }
    }
    limiter(key, mode, interval, limit, ts) {
        try {
            if (!this.operationTimeout) {
                return this.redis.limiter(key, mode, interval, limit, ts);
            }
            else {
                return this.promiseWithTimeout(this.redis.limiter(key, mode, interval, limit, ts));
            }
        }
        catch (e) {
            return Promise.reject(e);
        }
    }
    promiseWithTimeout(operationPromise) {
        let timer;
        const timeoutPromise = new Promise((_resolve, reject) => {
            timer = setTimeout(() => {
                reject(new Error('Operation timed out'));
            }, this.operationTimeout);
        });
        return Promise.race([operationPromise, timeoutPromise]).then((result) => {
            if (timer)
                clearTimeout(timer);
            return result;
        });
    }
}
exports.RedisSlidingWindowRateLimiter = RedisSlidingWindowRateLimiter;
exports.default = RedisSlidingWindowRateLimiter;
